/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../../state/reducers";
import {Checkbox, Icon, Table} from "semantic-ui-react";
import {
  getAssessmentStatisticsSiteToLinkPath,
  getSubmissionDownloadLink,
  getTutorViewForSubmission
} from "../../../../constants";
import {AssessmentFullBase} from "../../../../types/submissions";
import {
  rerunSingleSubmissionAsync,
  setActivePaginationPage,
  setPaginationPageSize,
  setRunningAssessment, setShowLastEditingIpAddress,
  setSortByKey
} from "../../../../state/actions/ownOrGroupExerciseActions/submissionsSite/submissionsSiteActions";
import {isCurrentSortingColumnPaginated, rotateSorting} from "../../../../helpers/tablesHelper";
import {HelpPopup} from "../../../helpers/helpPopup";
import {Link} from "react-router-dom";
import {isNotEmpty} from "../../../../validation/validations";


import {ArrayHelper} from "../../../../helpers/arrayHelper";
import {displayNumber} from "../../../../helpers/convertersAndTransformers";
import {MyPopup} from "../../../helpers/myPopup";
import {getI18n} from "../../../../../i18n/i18nRoot";
import PaginationTableFooter from '../../../helpers/paginationTableFooter'
import {askDialog} from '../../../../helpers/dialogHelper'

//const css = require('./styles.styl');

export interface MyProps {
  //readonly test: string
}

const mapStateToProps = (rootState: RootState /*, props: MyProps*/) => {
  return {
    //test0: rootState...
    //test: props.test
    release: rootState.submissionsSiteState.release,
    selectedReleaseId: rootState.submissionsSiteState.selectedReleaseId,
    submissions: rootState.submissionsSiteState.submissions,
    comesFromOwnExercisesSite: rootState.submissionsSiteState.comesFromOwnExercisesSite,
    pLangs: rootState.pLangsState.pLangs,
    runningAssessment: rootState.submissionsSiteState.runningAssessment,

    showLastEditingIpAddress: rootState.submissionsSiteState.showLastEditingIpAddress,

    sortByKey: rootState.submissionsSiteState.pagination.orderByProp,
    sortDirection: rootState.submissionsSiteState.pagination.isOrderByDesc,
    pagination: rootState.submissionsSiteState.pagination,
    langId: rootState.i18nState.langId
  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here
                                                                        rerunSingleSubmissionAsync,
                                                                        setRunningAssessment,
                                                                        setSortByKey,
                                                                        setPaginationPageSize,
                                                                        setActivePaginationPage,
                                                                        setShowLastEditingIpAddress,
                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;

class ListView extends React.Component<Props, any> {
  render(): JSX.Element {

    const selectedRelease = this.props.release

    if (!selectedRelease) {
      //TODO only error if really not found
      //on first render this might not be set because the list was not yet set
      return null
    }

    const canRerunAssessment =
      selectedRelease.hasAutomaticAssessmentStarted
      && selectedRelease.hasAutomaticAssessmentFinished

    const list: ReadonlyArray<AssessmentFullBase> = this.props.submissions

    //calculate passed percentages (normal, submit)

    // noinspection TsLint
    let passedAllNormalTestsPercentage: number | null = null
    // noinspection TsLint
    let passedAllNormalTestsCount: number | null = null
    // noinspection TsLint
    let passedAllSubmitTestsPercentage: number | null = null
    // noinspection TsLint
    let passedAllSubmitTestsCount: number | null = null

    const userCountWithNormalTestPoints = ArrayHelper.count(list,
                                                            p => p.normalTestPoints !== null
    )
    const userCountWithSubmitTestPoints = ArrayHelper.count(list,
                                                            p => p.submitTestPoints !== null
    )

    if (canRerunAssessment && list.length > 0) {
      passedAllNormalTestsCount = ArrayHelper.count(list,
                                                    (
                                                      el => el.normalTestPoints !== null && el.normalTestPoints === el.maxNormalTestPoints && el.shouldNotCount === false)
      )


      // noinspection TsLint
      passedAllNormalTestsPercentage = userCountWithNormalTestPoints !== 0
                                       ? passedAllNormalTestsCount * 100 / userCountWithNormalTestPoints
                                       : null


      passedAllSubmitTestsCount = ArrayHelper.count(list,
                                                    (
                                                      el => el.submitTestPoints !== null && el.submitTestPoints === el.maxSubmitTestPoints && el.shouldNotCount === false)
      )

      // noinspection TsLint
      passedAllSubmitTestsPercentage = userCountWithSubmitTestPoints !== 0
                                       ? passedAllSubmitTestsCount * 100 / userCountWithSubmitTestPoints
                                       : null
    }

    // noinspection TsLint
    let percentageShouldNotCount: number = null

    // noinspection TsLint
    let shouldNotCountCount: number = null
    if (list.length > 0) {
      shouldNotCountCount = ArrayHelper.count(list, el => el.shouldNotCount)
      // noinspection TsLint
      percentageShouldNotCount = shouldNotCountCount * 100 / list.length
    }

    return (
      <div>

        <Table selectable sortable padded striped singleLine className="material-table">
          <Table.Header>
            <Table.Row>
              <Table.HeaderCell>
                <span>
                  {
                    '#'
                  }
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'lastName',
                                                                      this.props.sortDirection
                  )
                }
              >
              <span className="clickable"
                    onClick={() => {
                      this.props.setSortByKey('lastName',
                                              rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                                this.props.sortByKey,
                                                                                'lastName'
                                              )
                      )
                    }}
              >
                {
                  getI18n(this.props.langId, 'Last name')
                }
              </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'firstName',
                                                                      this.props.sortDirection
                  )
                }
              >
                <span
                  className="clickable"
                  onClick={() => {
                    this.props.setSortByKey('firstName',
                                            rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                              this.props.sortByKey,
                                                                              'firstName'
                                            )
                    )
                  }}
                >
                  {
                    getI18n(this.props.langId, 'First name')
                  }
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'pLangId',
                                                                      this.props.sortDirection
                  )
                }
              >
                <span
                  className="clickable"
                  onClick={() => {
                    this.props.setSortByKey('pLangId',
                                            rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                              this.props.sortByKey,
                                                                              'pLangId'
                                            )
                    )
                  }}
                >
                  <HelpPopup icon="puzzle" defaultText={getI18n(this.props.langId, 'Programming language')}/>
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'shouldNotCount',
                                                                      this.props.sortDirection
                  )
                }
              >
                <span
                  className="clickable"
                  onClick={() => {
                    this.props.setSortByKey('shouldNotCount',
                                            rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                              this.props.sortByKey,
                                                                              'shouldNotCount'
                                            )
                    )
                  }}
                >
                  <HelpPopup icon="flag outline"
                             defaultText={getI18n(this.props.langId, 'Should not count')}/>
                  {
                    shouldNotCountCount !== null &&
                    <MyPopup
                      trigger={
                        <span>
                          {
                            percentageShouldNotCount !== null &&
                            ` (${displayNumber(percentageShouldNotCount, 2)}%)`
                          }
                        </span>
                      }
                      content={
                        <div>
                          {shouldNotCountCount}/{list.length}
                        </div>
                      }
                    />
                  }
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'participationCount',
                                                                      this.props.sortDirection
                  )
                }
              >
                <span
                  className="clickable"
                  onClick={() => {
                    this.props.setSortByKey('participationCount',
                                            rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                              this.props.sortByKey,
                                                                              'participationCount'
                                            )
                    )
                  }}
                >
                  <HelpPopup icon="repeat" wide
                             defaultText={getI18n(this.props.langId,
                                                  "The times the user solved this exercise (e.g. in different releases). If a user solved the exercise (release) in different programming languages then two rows are shown but it's the same participation. The participation is counted as one because the restrictions (e.g. lef time, should not count) are stored in the participation."
                             )}/>
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'normalTestPoints',
                                                                      this.props.sortDirection
                  )
                }
              >

                <span
                  className="clickable mar-right-half"
                  onClick={() => {
                    this.props.setSortByKey('normalTestPoints',
                                            rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                              this.props.sortByKey,
                                                                              'normalTestPoints'
                                            )
                    )
                  }}
                >
                  <HelpPopup wide icon="flag"
                             defaultText={getI18n(this.props.langId,
                                                  'The summed normal test points. The percentage represent the students where all normal tests passed and the participation should count. If nothing is displayed none normal test was run yet.'
                             )}/>

                  {
                    passedAllNormalTestsCount !== null &&
                    <MyPopup
                      trigger={
                        <span>
                          {
                            passedAllNormalTestsPercentage !== null &&
                            ` (${displayNumber(passedAllNormalTestsPercentage, 2)}%)`
                          }
                        </span>
                      }
                      content={
                        <div>
                          {passedAllNormalTestsCount}/{list.length} {` ${getI18n(this.props.langId,
                                                                                 "with all points"
                        )}`}
                        </div>
                      }
                    />
                  }
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'submitTestPoints',
                                                                      this.props.sortDirection
                  )
                }
              >
                <span className="clickable mar-right-half"
                      onClick={() => {
                        this.props.setSortByKey('submitTestPoints',
                                                rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                                  this.props.sortByKey,
                                                                                  'submitTestPoints'
                                                )
                        )
                      }}
                >
                  <HelpPopup wide icon="flag checkered"
                             defaultText={getI18n(this.props.langId,
                                                  'The summed submit test points. The percentage represent the students where all submit tests passed and the participation should count. If nothing is displayed none submit test was run yet.'
                             )}/>

                  {
                    passedAllSubmitTestsCount !== null &&
                    <MyPopup
                      trigger={
                        <span>
                          {
                            passedAllSubmitTestsPercentage !== null &&
                            ` (${displayNumber(passedAllSubmitTestsPercentage, 2)}%)`
                          }
                        </span>
                      }
                      content={
                        <div>
                          {passedAllSubmitTestsCount}/{list.length} {` ${getI18n(this.props.langId,
                                                                                 "with all points"
                        )}`}
                        </div>
                      }
                    />
                  }
                </span>
              </Table.HeaderCell>
              <Table.HeaderCell
                sorted={
                  isCurrentSortingColumnPaginated<AssessmentFullBase>(this.props.sortByKey, 'manualPoints',
                                                                      this.props.sortDirection
                  )
                }
              >
                <span className="clickable mar-right-half"
                      onClick={() => {
                        this.props.setSortByKey('manualPoints',
                                                rotateSorting<AssessmentFullBase>(this.props.sortDirection,
                                                                                  this.props.sortByKey,
                                                                                  'manualPoints'
                                                )
                        )
                      }}
                >
                  {
                    getI18n(this.props.langId, 'Manual')
                  }
                </span>
                <HelpPopup wide
                           defaultText={getI18n(this.props.langId, 'Points that a tutor has assigned')}/>
              </Table.HeaderCell>
              <Table.HeaderCell>
                <HelpPopup icon="sticky note"
                           defaultText={getI18n(this.props.langId, 'Note for other tutors')}/>
              </Table.HeaderCell>
              <Table.HeaderCell>
                <span className="mar-right-half">
                  {
                    getI18n(this.props.langId, 'Open')
                  }
                </span>
                <HelpPopup
                  wide
                  defaultText={getI18n(this.props.langId,
                                       'In the tutor view you can view the submission, the test results and manually assess the submission.'
                  )}/>
              </Table.HeaderCell>
              <Table.HeaderCell>

                {
                  /*
                  <Icon name="sticky note" />
                <HelpPopup defaultText={getI18n(this.props.langId,'Feedback for student')}/>
                   */
                }
                <HelpPopup icon="talk" defaultText={getI18n(this.props.langId, 'Feedback for student')}/>

              </Table.HeaderCell>
              <Table.HeaderCell>

                <HelpPopup className="mar-left-half" icon="download"
                           defaultText={getI18n(this.props.langId, 'Download')}/>
                <HelpPopup icon="exclamation triangle" defaultText={getI18n(this.props.langId,
                                                                            'There is an option issue with this feature/function:'
                ) + '#19 Server'}/>
              </Table.HeaderCell>
              <Table.HeaderCell>
                <HelpPopup icon="area graph" defaultText={getI18n(this.props.langId, 'Assessment statistic')}/>
              </Table.HeaderCell>

              <Table.HeaderCell>
                <HelpPopup icon="bug" defaultText={getI18n(this.props.langId,
                                                           'Error during assessment (before or after testing, e.g. test server is offline)'
                )}/>
              </Table.HeaderCell>

              <Table.HeaderCell>
                <span className={this.props.showLastEditingIpAddress
                                 ? 'clickable-toggled'
                                 : 'clickable'} onClick={() => {
                  this.props.setShowLastEditingIpAddress(!this.props.showLastEditingIpAddress)
                }}>
                  <HelpPopup icon="desktop"
                             defaultText={`${getI18n(this.props.langId, "Last editing ip address")}. ${getI18n(
                               this.props.langId, "Click to expand all")}`}/>
                </span>
              </Table.HeaderCell>

              <Table.HeaderCell>
                {
                  getI18n(this.props.langId, 'Re-run')
                }
              </Table.HeaderCell>
            </Table.Row>
          </Table.Header>

          <Table.Body>
            {
              list.map((submission: AssessmentFullBase, index) => {
                return (
                  <Table.Row key={`${submission.userId}-${submission.pLangId}`}>
                    <Table.Cell>
                      <div className={
                        this.props.runningAssessment
                        && this.props.runningAssessment.pLangId === submission.pLangId
                        && this.props.runningAssessment.userId === submission.userId
                        ? `table-line-active`
                        : null
                      }>
                        {
                          ((this.props.pagination.currentPage - 1) * this.props.pagination.pageSize + index + 1)
                        }
                      </div>
                    </Table.Cell>
                    <Table.Cell>
                      <div>
                        {submission.lastName}
                      </div>
                    </Table.Cell>
                    <Table.Cell>
                      <span>
                        {submission.firstName}
                      </span>
                    </Table.Cell>
                    <Table.Cell>
                      <span>
                        {
                          this.props.pLangs.find(p => p.id === submission.pLangId) &&
                          this.props.pLangs.find(p => p.id === submission.pLangId).displayName
                        }
                      </span>
                    </Table.Cell>
                    <Table.Cell>
                      <div>
                        <Checkbox checked={submission.shouldNotCount} disabled/>
                      </div>
                    </Table.Cell>
                    <Table.Cell>
                      <span>
                        {
                          submission.participationCount === -1
                          ? getI18n(this.props.langId,
                                    'Error'
                          )
                          : submission.participationCount
                        }
                      </span>
                    </Table.Cell>
                    <Table.Cell>
                      <span>
                        {
                          submission.normalTestPoints !== null &&
                          `${submission.normalTestPoints} / ${submission.maxNormalTestPoints}`
                        }
                      </span>
                    </Table.Cell>

                    <Table.Cell>
                      <span>
                        {
                          submission.submitTestPoints !== null &&
                          `${submission.submitTestPoints} / ${submission.maxSubmitTestPoints}`
                        }
                      </span>
                    </Table.Cell>

                    <Table.Cell>
                      {
                        submission.manualPoints !== null &&
                        <span>
                          {
                            submission.manualPoints
                          }
                          {
                            this.props.release.maxManualPoints !== null &&
                            <span>
                              {
                                `/${this.props.release.maxManualPoints}`
                              }
                            </span>
                          }
                          </span>
                      }

                    </Table.Cell>

                    <Table.Cell>
                      {
                        isNotEmpty(submission.noteForOtherTutors) &&
                        <HelpPopup defaultText={
                          <pre className="notes-popup">
                            {
                              submission.noteForOtherTutors
                            }
                          </pre>
                        }
                                   style={{margin: '0'}}
                                   wide='very'
                        />
                      }
                    </Table.Cell>

                    <Table.Cell>
                      <div>
                        <Link to={getTutorViewForSubmission(this.props.comesFromOwnExercisesSite,
                                                            this.props.selectedReleaseId, submission.userId,
                                                            submission.pLangId
                        )}>
                          <Icon name="life ring"/>
                        </Link>
                      </div>
                    </Table.Cell>

                    <Table.Cell>
                      {
                        isNotEmpty(submission.feedbackForStudent) &&
                        <HelpPopup defaultText={
                          <pre className="notes-popup">
                            {
                              submission.feedbackForStudent
                            }
                          </pre>
                        }
                                   style={{margin: '0'}}
                                   wide='very'
                        />
                      }
                    </Table.Cell>

                    <Table.Cell>
                      <div>
                        <a target="_blank"
                           download
                           href={
                             getSubmissionDownloadLink(
                               (selectedRelease
                                ? selectedRelease.generatedCode
                                : null),
                               submission.userId,
                               submission.pLangId
                             )
                           }>
                          <Icon name="download"/>
                        </a>
                      </div>
                    </Table.Cell>

                    <Table.Cell>
                      <div>
                        <Link style={{margin: '0'}} to={getAssessmentStatisticsSiteToLinkPath(submission.userId)}>
                          <Icon name="area graph"/>
                        </Link>
                      </div>
                    </Table.Cell>

                    <Table.Cell>
                      {
                        submission.lastAssessmentErrorMessage &&
                        <div>
                          <HelpPopup icon="bug" defaultText={
                            <pre className="notes-popup">
                            {
                              submission.lastAssessmentErrorMessage
                            }
                          </pre>
                          }
                                     style={{margin: '0'}}
                                     wide='very'
                          />
                        </div>
                      }
                    </Table.Cell>

                    <Table.Cell>
                      {
                        (this.props.showLastEditingIpAddress && submission.lastEditingIpAddress) &&
                        <span>
                        {
                          submission.lastEditingIpAddress
                        }
                        </span>
                      }

                      {
                        (this.props.showLastEditingIpAddress === false && submission.lastEditingIpAddress) &&
                        <HelpPopup defaultText={submission.lastEditingIpAddress}
                                   style={{margin: '0'}}
                        />
                      }

                    </Table.Cell>
                    <Table.Cell>
                      <div className={canRerunAssessment
                                      ? 'clickable'
                                      : ' div-disabled'}
                           onClick={async () => {

                             const shouldRerun = await askDialog(getI18n(this.props.langId, "Re-run"),
                                                                 getI18n(this.props.langId,
                                                                         "Are you sure you want to re-run the automatic assessment for this student?"
                                                                 ), this.props.langId
                             )

                             if (!shouldRerun) return

                             this.props.setRunningAssessment(submission)

                             await this.props.rerunSingleSubmissionAsync(
                               selectedRelease
                               ? selectedRelease.generatedCode
                               : null, submission, true)

                             this.props.setRunningAssessment(null)
                           }}


                      >
                        <Icon name="refresh"/>
                      </div>
                    </Table.Cell>

                  </Table.Row>
                )
              })
            }

          </Table.Body>
          <PaginationTableFooter
            pagination={this.props.pagination}
            setPaginationPageSize={this.props.setPaginationPageSize}
            setActivePaginationPage={this.props.setActivePaginationPage}
          />
        </Table>

      </div>
    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(ListView)
