/**
 * Created by janis dähne.
 */
import {CREATE_customProjectAction} from "../../reducers/editCustomProjectSite/crud/createCustomProjectReducer";
import {
  CustomProjectDescriptionFullBase,
  CustomProjectFrontendExclusive,
  CustomProjectFullBase,
  CustomProjectPropertiesFromFrontend,
  CustomProjectSolutionFileFullBase,
  CustomProjectSolutionFullBase,
  RunnableCustomProjectTestFrontendExclusive
} from "../../../types/customProjects";
import {ActionType} from "../../reducers/editCustomProjectSite/editCustomProjectSiteActionTypes";
import {delay} from "../../../../debug";
import {
  attachExitingFileToCustomProjectCustomTest, attachExitingFileToCustomProjectDescription,
  createCustomProject, detachFileFromCustomProjectCustomTest, detachFileFromCustomProjectDescription,
  getCustomProject,
  getOwnCustomProjectTestResults, getSingleCustomProjectDescriptionAssetFile,
  updateCustomProjectDescription,
  updateCustomProjectProperties,
  updateCustomProjectSolution
} from "../../../communicationLayer/editCustomProjectLayer";
import {GET_customProjectAction} from "../../reducers/editCustomProjectSite/crud/getCustomProjectReducer";
import {createAndRefreshAssetUrlsAndObjects} from "../../../helpers/assetHelper";
import {AwaitActions, AwaitActionsWithResult} from "../types";
import {UPDATE_customProjectDescriptionAction} from "../../reducers/editCustomProjectSite/crud/updateCustomProjectDescriptionReducer";
import {UPDATE_customProjectSolutionAction} from "../../reducers/editCustomProjectSite/crud/updateCustomProjectSolutionReducer";
import {
  setHasFileChangedMapCustomProjectSolutionEditor, setHasUnsavedChanges
} from "./editorActions/customProjectSolutionEditorActions";
import {resetCustomProjectTestResultsDuringSave, setSelectedConsoleViewTest} from "./editCustomProjectActions";
import Logger from "../../../helpers/logger";
import isEqual from 'lodash-es/isEqual'
import orderBy from 'lodash-es/orderBy'
import {
  JustRunProgramTestCustomProjectCommandForBackend,
  RunCompileSingleFileCommandCustomProjectForBackend,
  RunTestCommandCustomProjectForBackend
} from "../../../types/testsAndTestProtocol";
import {KnownInternalTestTypes, TestTypeFullBase} from "../../../types/testTypes";
import {errorDialog, warningDialog} from "../../../helpers/dialogHelper";
import {RUN_compileSingleFileCustomProjectAction} from "../../reducers/editCustomProjectSite/crud/runCompileSingleFileCustomProjectReducer";
import {
  compileSingleFileCustomProject, justRunProgramTestCustomProject, runCustomTestCustomProject
} from "../../../communicationLayer/testingLayer";
import {RUN_customProjectTestAction} from "../../reducers/editCustomProjectSite/crud/runCustomProjectTestReducer";
import {setCustomProjectTestIsRunning} from "./subSet/customProjectTestActions";
import {GET_customProjectTestResultsAction} from "../../reducers/editCustomProjectSite/crud/getCustomProjectTestResultsReducer";
import {UPDATE_customProjectPropertiesAction} from "../../reducers/editCustomProjectSite/crud/updateCustomProjectProperties";
import {JUST_RUN_programAction} from "../../reducers/editCustomProjectSite/crud/justRunProgramReducer";
import {RunnableExerciseTestFrontendExclusive} from "../../../types/exerciseEditor";
import {getI18n, KnownLangs} from "../../../../i18n/i18nRoot";
import globalState from '../../state'
import {
  AceEditorsStorageHelper,
  AceStoredTabLike
} from '../../../components/codeEditors/aceEditors/aceEditorsStorageHelper'
import {customProjectEditorPanelViewEditorId, doEditorEditorPanelViewEditorId} from '../../../constants'
import {ATTACH_customProjectCustomTestAssetPreviewEAction} from '../../reducers/editCustomProjectSite/crud/assets/attachCustomProjectCustomTestAssetPreviewReducer'
import {DETACH_customProjectCustomTestAssetPreviewAction} from '../../reducers/editCustomProjectSite/crud/assets/detachCustomProjectCustomTestAssetPreviewReducer'
import {ATTACH_customProjectMarkdownAssetPreviewAction} from '../../reducers/editCustomProjectSite/crud/assets/attachCustomProjectMarkdownAssetPreviewReducer'
import {DETACH_customProjectMarkdownAssetPreviewAction} from '../../reducers/editCustomProjectSite/crud/assets/detachCustomProjectMarkdownAssetPreviewReducer'
import {LOAD_customProjectMarkdownAssetAction} from '../../reducers/editCustomProjectSite/crud/assets/loadSingleCustomProjectMarkdownAssetReducer'


export function updateCustomProjectPropertiesAsync(customProjectId: number,
                                                   customProjectProperties: CustomProjectPropertiesFromFrontend
): UPDATE_customProjectPropertiesAction {
  return {
    type: ActionType.UPDATE_customProjectProperties,
    payload: delay(updateCustomProjectProperties(customProjectId, customProjectProperties))
  }
}


export function getCustomProjectTestResultsAsync(customProjectId: number,
                                                 pLangId: number
): GET_customProjectTestResultsAction {
  return {
    type: ActionType.GET_customProjectTestResults,
    payload: delay(getOwnCustomProjectTestResults(customProjectId, pLangId))
  }
}

/**
 *
 * copied from doExerciseCrudActions > runCustomDoExerciseTest
 * @param {RunTestCommandForBackend} runNormalTestCommand
 * @param langId
 * @returns {AwaitActionsWithResult<boolean>} true: all tests passed, false: at least 1 did not compile or not passed
 */
export function runCustomProjectTestAsync(runNormalTestCommand: RunTestCommandCustomProjectForBackend, langId: KnownLangs): AwaitActionsWithResult<boolean> {
  return async (dispatch, getState) => {

    //save changes if code changed
    if (getState().editCustomProjectSiteState.solutionEditorState.hasUnsavedChanges) {
      //first save changes
      //normally this should be func arguments but would the component re render on every code change??
      const hasSaved = await dispatch(
        saveCustomProjectSolution(getState().editCustomProjectSiteState.customProject, langId))

      if (!hasSaved) {
        return false
      }
    }

    //if we run the same test twice and the output won't change then this looks like nothing happened...
    //set console to blank so that the user sees that something happened
    dispatch(setSelectedConsoleViewTest(null))

    const runningTests = runNormalTestCommand.testIds
      .map<RunnableCustomProjectTestFrontendExclusive>(testId =>
                                                         //don't use allTests because saveDoExerciseSolution could have reset the test (results) and allTest would be an old array
                                                         getState().editCustomProjectSiteState.customProject.tests
                                                           .find(test =>
                                                                   test.id === testId
                                                           )
      )

    //set test as running to show animation
    dispatch(setCustomProjectTestIsRunning(runningTests, getState().editCustomProjectSiteState.customProject.tests))
    await dispatch(_runCustomTestCustomProjectAsync(runNormalTestCommand))

    const updatedTest = getState().editCustomProjectSiteState.customProject.tests.find(
      p => p.id === runNormalTestCommand.testIds[runNormalTestCommand.testIds.length - 1])

    if (!updatedTest) {
      warningDialog(
        getI18n(globalState.getState().i18nState.langId, 'Warning'),
        getI18n(globalState.getState().i18nState.langId, 'Probably no test has run'), globalState.getState().i18nState.langId
      )
      return
    }

    const ranTests = runNormalTestCommand.testIds
      .map<RunnableCustomProjectTestFrontendExclusive>(testId =>
                                                         getState().editCustomProjectSiteState.customProject.tests
                                                           .find(test =>
                                                                   test.id === testId
                                                           )
      )

    const firstFailedTest = ranTests.find(p => !p || !p.result || !p.result.hasCompiled || !p.result.passed)

    if (firstFailedTest) {
      //select the first failed test
      dispatch(setSelectedConsoleViewTest(firstFailedTest))
    } else {
      //select the last test
      dispatch(setSelectedConsoleViewTest(updatedTest))
    }

    return !!firstFailedTest
  }
}


export function _runCustomTestCustomProjectAsync(runNormalTestCommand: RunTestCommandCustomProjectForBackend): RUN_customProjectTestAction {
  return {
    type: ActionType.RUN_customProjectTest,
    payload: delay(runCustomTestCustomProject(runNormalTestCommand))
  }
}

export function runCompileSingleFileCustomProjectAsync(runCompileSingleFileCommand: RunCompileSingleFileCommandCustomProjectForBackend,
                                                       testTypes: ReadonlyArray<TestTypeFullBase>
): AwaitActions {
  return async (dispatch, getState) => {

    const compileTestType = testTypes.find(p => p.internalName === KnownInternalTestTypes.compileTest)

    if (!compileTestType) {

      errorDialog(getI18n(globalState.getState().i18nState.langId, 'Error'),
                  getI18n(globalState.getState().i18nState.langId, 'Unknown test type'),globalState.getState().i18nState.langId
      )

      return
    }

    await dispatch(_runCompileSingleFileCustomProject(runCompileSingleFileCommand))

    const compileResult = getState().editCustomProjectSiteState.lastCompileSingleFileResult

    const tempTest: RunnableCustomProjectTestFrontendExclusive = {
      id: -1,
      displayName: 'Compile Single File',
      result: compileResult,
      content: '',
      uiIsRunning: false,
      files: [],
      displayIndex: -1,
      weight: 0,
      uiIsAfterTest: false,
      testSettings: {
        id: -1,
        maxDiskSpaceInKb: -1,
        timeoutInMs: -1,
        memoryLimitInKb: -1,
        compileTimeoutInMs: -1
      },
      testTypeId: compileTestType.id
    }

    dispatch(setSelectedConsoleViewTest(tempTest))

  }
}


export function _runCompileSingleFileCustomProject(runCompileSingleFileCommand: RunCompileSingleFileCommandCustomProjectForBackend): RUN_compileSingleFileCustomProjectAction {
  return {
    type: ActionType.RUN_compileSingleFileCustomProject,
    payload: delay(compileSingleFileCustomProject(runCompileSingleFileCommand))
  }
}


export function justRunProgramCustomProjectAsync(justRunProgramCommand: JustRunProgramTestCustomProjectCommandForBackend,
                                                 testTypes: ReadonlyArray<TestTypeFullBase>
): AwaitActions {
  return async (dispatch, getState) => {

    // const justRunTestType = testTypes.find(p => p.internalName === KnownInternalTestTypes.justRunTest)
    //
    // if (!justRunTestType) {
    //
    //   errorDialog(getI18n(globalState.getState().i18nState.langId,'Error'),
    //     getI18n(globalState.getState().i18nState.langId,'Unknown test type'))
    //
    //   return
    // }

    await dispatch(_justRunProgramCustomProject(justRunProgramCommand))

    const testResult = getState().editCustomProjectSiteState.lastJustRunProgramResult

    const tempTest: RunnableExerciseTestFrontendExclusive = {
      id: -1,
      displayName: 'Just Run Program',
      result: testResult,
      content: '',
      uiIsRunning: false,
      files: [],
      displayIndex: -1,
      weight: 0,
      isSubmitTest: false,
      uiIsAfterTest: false,
      testSettings: {
        id: -1,
        maxDiskSpaceInKb: -1,
        timeoutInMs: -1,
        memoryLimitInKb: -1,
        compileTimeoutInMs: -1,
      },
      testTypeId: -1
    }

    dispatch(setSelectedConsoleViewTest(tempTest))

  }
}

export function _justRunProgramCustomProject(justRunProgramCommand: JustRunProgramTestCustomProjectCommandForBackend): JUST_RUN_programAction {
  return {
    type: ActionType.JUST_RUN_program,
    payload: delay(justRunProgramTestCustomProject(justRunProgramCommand))
  }
}


export function getCustomProjectAsync(customProjectId: number): AwaitActions {
  return async (dispatch, getState) => {

    try {
      await dispatch(_getCustomProjectAsync(customProjectId))
    } catch (err) {
      return
    }

    //create asset urls
    createAndRefreshAssetUrlsAndObjects(getState().editCustomProjectSiteState.customProject.description.assets)
  }
}

function _getCustomProjectAsync(customProjectId: number): GET_customProjectAction {
  return {
    type: ActionType.GET_customProject,
    payload: delay(getCustomProject(customProjectId))
  }
}

export function createCustomProjectAsync(customProject: CustomProjectFullBase): AwaitActionsWithResult<boolean> {
  return async (dispatch, getState) => {

    try {
      await dispatch(_createCustomProjectAsync(customProject))
    } catch (err) {
      Logger.error(`could not create new custom project, err: ${err.message}`)
      return false
    }

    return true
  }
}

export function _createCustomProjectAsync(customProject: CustomProjectFullBase): CREATE_customProjectAction {
  return {
    type: ActionType.CREATE_customProject,
    payload: delay(createCustomProject(customProject))
  }
}


function needToSaveSolution(backendSolution: CustomProjectSolutionFullBase,
                            frontendSolution: CustomProjectSolutionFullBase
): boolean {

  const backendSolutionVersion = JSON.stringify(backendSolution,
                                                (key: string, value: any) => {
                                                  if (key === 'isDisplayed') {
                                                    return 'true'
                                                  }
                                                  if (key === 'displayIndex') {
                                                    return '0'
                                                  }
                                                  return value
                                                }
  )

  const frontendSolutionVersion = JSON.stringify(frontendSolution,
                                                 (key: string, value: any) => {
                                                   if (key === 'isDisplayed') {
                                                     return 'true'
                                                   }
                                                   if (key === 'displayIndex') {
                                                     return '0'
                                                   }
                                                   return value
                                                 }
  )

  // noinspection TsLint
  let cleanedBackendVersion: CustomProjectSolutionFullBase = JSON.parse(backendSolutionVersion)
  // noinspection TsLint
  let cleanedFrontendVersion: CustomProjectSolutionFullBase = JSON.parse(frontendSolutionVersion)


  //we need this because backend does not include edit sessions...
  cleanedFrontendVersion = {
    ...cleanedFrontendVersion,
    solutionFiles: orderBy(cleanedFrontendVersion.solutionFiles,
                           (p: CustomProjectSolutionFileFullBase) => p.id
    ).map<CustomProjectSolutionFileFullBase>(p => {
      return {
        ...p,
      }
    })
  }

  cleanedBackendVersion = {
    ...cleanedBackendVersion,
    solutionFiles: orderBy(cleanedBackendVersion.solutionFiles,
                           (p: CustomProjectSolutionFileFullBase) => p.id
    ).map<CustomProjectSolutionFileFullBase>(p => {
      return {
        ...p,
      }
    })
  }

  const areEqual = isEqual(cleanedFrontendVersion, cleanedBackendVersion)


  return !areEqual

}


/**
 *
 * @param {CustomProjectFullBase} customProject
 * @param langId
 * @returns {AwaitActionsWithResult<boolean>} true: all ok, false: not
 */
export function saveCustomProjectSolution(customProject: CustomProjectFullBase, langId: KnownLangs): AwaitActionsWithResult<boolean> {
  return async (dispatch, getState) => {

    //we call this because ... see MultiFileAceEditor

    try {
      await dispatch(_updateCustomProjectSolutionAsync(customProject.id, customProject.solution, langId))
    } catch (err) {
      return false
    }

    dispatch(resetCustomProjectTestResultsDuringSave())

    dispatch(setHasUnsavedChanges(false))
    dispatch(setHasFileChangedMapCustomProjectSolutionEditor({}))

    return true
  }
}

/**
 *
 * @param {CustomProjectFullBase} customProject
 * @param langId
 * @returns {AwaitActionsWithResult<boolean>} true: all ok, false: some error
 */
export function updateCustomProjectAsync(customProject: CustomProjectFrontendExclusive, langId: KnownLangs): AwaitActions {
  return async (dispatch, getState) => {

    //this is much faster than shouldSaveSolution so check this first
    const solutionEditorHasUnsavedChanges = getState().editCustomProjectSiteState.solutionEditorState.hasUnsavedChanges

    // noinspection TsLint
    let shouldSaveSolution = false

    if (solutionEditorHasUnsavedChanges === false) {
      shouldSaveSolution = needToSaveSolution(getState().editCustomProjectSiteState.initialSolutionFromBackend,
                                              customProject.solution
      )
    }

    if (solutionEditorHasUnsavedChanges || shouldSaveSolution) {

      Logger.debug('editCustomProjectCrudActions', 'saving solution')

      await dispatch(saveCustomProjectSolution(customProject, langId))

    } else {
      Logger.debug('editCustomProjectCrudActions', 'NOT saving solution because equal')
    }

    //update the ace storage

    const pLang = getState().pLangsState.pLangs.find(
      p => p.id === getState().editCustomProjectSiteState.initialSolutionFromBackend.pLangId)

    if (!pLang) {
      Logger.error('could not find plang')
    }

    AceEditorsStorageHelper.removeEditor(customProjectEditorPanelViewEditorId)

    const storedTabs = getState().editCustomProjectSiteState.initialSolutionFromBackend.solutionFiles.map<AceStoredTabLike>(
      p => {
        return {
          id: p.id,
          content: p.content,
          language: pLang.internalName
        }
      })
    AceEditorsStorageHelper.addEditorWithTabs(customProjectEditorPanelViewEditorId, storedTabs)

    const equalDescription = isEqual(getState().editCustomProjectSiteState.initialDescriptionFromBackend,
                                     customProject.description
    )

    if (equalDescription) {
      //backend already have the latest version
      Logger.debug('editCustomProjectCrudActions', 'NOT saving description because equal')
    } else {

      Logger.debug('editCustomProjectCrudActions', 'saving description')
      await dispatch(_updateCustomProjectDescriptionAsync(customProject.id, customProject.description))
    }


    Logger.debug('editCustomProjectCrudActions', 'saving properties')
    await dispatch(updateCustomProjectPropertiesAsync(customProject.id, {
      displayName: customProject.displayName
    }))


  }
}


export function _updateCustomProjectSolutionAsync(customProjectId: number,
                                                  customProjectSolution: CustomProjectSolutionFullBase,
                                                  langId: KnownLangs
): UPDATE_customProjectSolutionAction {
  return {
    type: ActionType.UPDATE_customProjectSolution,
    payload: delay(updateCustomProjectSolution(customProjectId, customProjectSolution)),
    meta: {
      langId
    }
  }
}

export function _updateCustomProjectDescriptionAsync(customProjectId: number,
                                                     customProjectDescription: CustomProjectDescriptionFullBase
): UPDATE_customProjectDescriptionAction {
  return {
    type: ActionType.UPDATE_customProjectDescription,
    payload: delay(updateCustomProjectDescription(customProjectId, customProjectDescription))
  }
}



export function attachExitingFileToCustomProjectDescriptionAsync(customProjectId: number, fileReferenceId: number): AwaitActions {
  return async (dispatch, getState) => {

    await dispatch(_attachExitingFileToCustomProjectDescriptionAsync(customProjectId, fileReferenceId))

    //after we connected the file to the exercise description...
    //we need to load it so it can be used!

    Logger.debug("editCustomProjectCrudActions", `after attaching file id: ${fileReferenceId} to custom project id: ${customProjectId}, we load the file content...`)

    await dispatch(_getSingleCustomProjectDescriptionAssetFileAsync(customProjectId, fileReferenceId))

  }
}

export function _attachExitingFileToCustomProjectDescriptionAsync(customProjectId: number, fileReferenceId: number): ATTACH_customProjectMarkdownAssetPreviewAction {
  return {
    type: ActionType.ATTACH_customProjectMarkdownAssetPreview,
    payload: delay(attachExitingFileToCustomProjectDescription(customProjectId, fileReferenceId))
  }
}


export function _getSingleCustomProjectDescriptionAssetFileAsync(customProjectId: number, fileReferenceId: number, onProgress?: (percentage: number) => void): LOAD_customProjectMarkdownAssetAction {
  return {
    type: ActionType.LOAD_customProjectMarkdownAsset,
    payload: delay(getSingleCustomProjectDescriptionAssetFile(customProjectId, fileReferenceId, onProgress))
  }
}


export function detachFileFromCustomProjectDescriptionAsync(customProjectId: number, fileReferenceId: number): DETACH_customProjectMarkdownAssetPreviewAction {
  return {
    type: ActionType.DETACH_customProjectMarkdownAssetPreview,
    payload: delay(detachFileFromCustomProjectDescription(customProjectId, fileReferenceId)),
    meta: {
      assetPreviewId: fileReferenceId
    }
  }
}


export function attachExitingFileToCustomProjectCustomTestAsync(customProjectId: number, customTestId: number, fileReferenceId: number): ATTACH_customProjectCustomTestAssetPreviewEAction {
  return {
    type: ActionType.ATTACH_customProjectCustomTestAssetPreview,
    payload: delay(attachExitingFileToCustomProjectCustomTest(customProjectId, customTestId, fileReferenceId)),
    meta: {
      customTestId
    }
  }
}

export function detachFileFromCustomProjectCustomTestAsync(customProjectId: number, customTestId: number, fileReferenceId: number): DETACH_customProjectCustomTestAssetPreviewAction {
  return {
    type: ActionType.DETACH_customProjectCustomTestAssetPreview,
    payload: delay(detachFileFromCustomProjectCustomTest(customProjectId, customTestId, fileReferenceId)),
    meta: {
      customTestId,
      assetPreviewId: fileReferenceId
    }
  }
}



