/**
 * Created by janis dähne.
 */
import {Action} from "redux";
import {notExhaustive} from "../../../_notExhausiveHelper";
import {ExerciseTestForBackend} from "../../../../../types/exerciseEditor";
import {
  AllActions as TestActions,
  reducer as testReducer,
  initial as initialTest,
  validationRules
} from './editorTestReducer'
import {ActionBase, ActionType} from "./changeTestDialogActionTypes";
import {isValidDialog} from "../../../../../validation/dialogValidations";
import {ActionTypeGlobal, RESET_GlobalAction} from '../../../globalActions'
import {ActionType as EditorExerciseSiteActionTypes} from '../../editorExerciseSiteActionTypes'

export type State = {
  readonly isBusy: boolean
  readonly isHelpDisplayed: boolean
  readonly isCreating: boolean

  readonly isValid: boolean

  readonly selectedTabIndex: number

  readonly test: ExerciseTestForBackend

}


export const initial: () => State = () => {
  return {

    isBusy: false,
    isHelpDisplayed: false,
    isCreating: false,
    selectedTabIndex: 0,

    test: initialTest(),
    isValid: isValidDialog(initialTest(), validationRules),
  }
}


export interface SET_selectedTabIndexAction extends ActionBase {
  readonly type: ActionType.SET_selectedTabIndex
  readonly index: number
}

export interface SET_isHelpDisplayedAction extends ActionBase {
  readonly type: ActionType.SET_isHelpDisplayed
  readonly isDisplayed: boolean
}

export interface SET_isCreatingAction extends ActionBase {
  readonly type: ActionType.SET_isCreating
  readonly isCreating: boolean
}


export interface ResetAction extends ActionBase {
  readonly type: ActionType.RESET
}


export type AllActions =
  ResetAction
  | SET_isHelpDisplayedAction
  | SET_isCreatingAction
  | SET_selectedTabIndexAction
  | TestActions

  | RESET_GlobalAction

export function reducer(state: State = initial(), action: AllActions): State {

  switch (action.type) {

    case ActionType.SET_testId:
    case ActionType.SET_content:
    case ActionType.SET_displayName:
    case ActionType.SET_displayIndex:
    case ActionType.SET_isSubmitTest:
    case ActionType.SET_weight:
    case ActionType.SET_testTypeId:
    case ActionType.SET_editorTest_memoryLimitInKb:
    case ActionType.SET_editorTest_maxDiskSpaceInKb:
    case ActionType.SET_editorTest_timeoutInMs:
    case ActionType.SET_editorTest_compileTimeoutInMs:
    case ActionType.SET_editorTest_files:
    case EditorExerciseSiteActionTypes.ATTACH_testAsset_FULFILLED:
    case EditorExerciseSiteActionTypes.DETACH_testAsset_FULFILLED:
      const temp = testReducer(state.test, action)
      return {
        ...state,
        isValid: isValidDialog(temp, validationRules),
        test: testReducer(state.test, action)
      }

    case ActionType.SET_isCreating:
      return {
        ...state,
        isCreating: action.isCreating
      }

    case ActionType.SET_isHelpDisplayed:
      return {
        ...state,
        isHelpDisplayed: action.isDisplayed
      }

    case ActionType.SET_selectedTabIndex:
      return {
        ...state,
        selectedTabIndex: action.index
      }

    case ActionType.RESET:
    case ActionTypeGlobal.RESET_Global:
      return initial()

    default:
      notExhaustive(action)
      return state
  }
}

