/**
 * Created by janis dähne.
 */

import {notExhaustive} from "../../../_notExhausiveHelper";
import {ExerciseTestForBackend, FilePreviewFromBackend} from "../../../../../types/exerciseEditor";
import {ActionBase, ActionType} from "./changeTestDialogActionTypes";
import {getValidationCollection, isNotEmptyOrSpaces, isValidId} from "../../../../../validation/validations";
import {getNextFrontendId} from "../../../../../helpers/frontendIdHelper";

import {ATTACH_testAsset_FULFILLEDAction} from '../../crud/assets/attachTestAssetPreviewReducer'
import {DETACH_testAsset_FULFILLEDAction} from '../../crud/assets/detachTestAssetPreviewReducer'
import {ActionType as EditorExerciseSiteActionTypes} from '../../editorExerciseSiteActionTypes'

export type State = ExerciseTestForBackend


//function to evaluate getNextFrontendId every time
export const initial: () => State = () => {
  return {
    id: getNextFrontendId(),
    content: '',
    displayName: '',
    displayIndex: -1, //use -1 and let determine the action the right index
    isSubmitTest: false,
    weight: 1,
    files: [],
    testTypeId: -1,
    //TODO maybe validation??
    testSettings: {
      id: getNextFrontendId(),
      memoryLimitInKb: 1000,
      timeoutInMs: 1000,
      maxDiskSpaceInKb: 1000,
      compileTimeoutInMs: 2000,
    }
  }
}


export const validationRules = getValidationCollection<ExerciseTestForBackend>({
                                                                                 id: [],
                                                                                 content: [],
                                                                                 displayName: [isNotEmptyOrSpaces],
                                                                                 displayIndex: [],
                                                                                 isSubmitTest: [],
                                                                                 weight: [],
                                                                                 files: [],
                                                                                 testTypeId: [isValidId],
                                                                                 testSettings: [],
                                                                               })

export interface SET_testIdAction extends ActionBase {
  readonly type: ActionType.SET_testId
  readonly id: number
}

export interface SET_contentAction extends ActionBase {
  readonly type: ActionType.SET_content
  readonly content: string
}

export interface SET_displayNameAction extends ActionBase {
  readonly type: ActionType.SET_displayName
  readonly displayName: string
}

export interface SET_displayIndexAction extends ActionBase {
  readonly type: ActionType.SET_displayIndex
  readonly displayIndex: number
}

export interface SET_isSubmitTestAction extends ActionBase {
  readonly type: ActionType.SET_isSubmitTest
  readonly isSubmitTest: boolean
}

export interface SET_weightAction extends ActionBase {
  readonly type: ActionType.SET_weight
  readonly weight: number
}

export interface SET_testTypeIdAction extends ActionBase {
  readonly type: ActionType.SET_testTypeId
  readonly testTypeId: number
}

export interface SET_editorTest_memoryLimitInKbAction extends ActionBase {
  readonly type: ActionType.SET_editorTest_memoryLimitInKb
  readonly memoryLimitInKb: number
}

export interface SET_editorTest_timeoutInMsAction extends ActionBase {
  readonly type: ActionType.SET_editorTest_timeoutInMs
  readonly timeoutInMs: number
}

export interface SET_editorTest_compileTimeoutInMsAction extends ActionBase {
  readonly type: ActionType.SET_editorTest_compileTimeoutInMs
  readonly compileTimeoutInMs: number
}

export interface SET_editorTest_maxDiskSpaceInKbAction extends ActionBase {
  readonly type: ActionType.SET_editorTest_maxDiskSpaceInKb
  readonly maxDiskSpaceInKb: number
}


export interface SET_editorTest_filesAction extends ActionBase {
  readonly type: ActionType.SET_editorTest_files
  readonly files: ReadonlyArray<FilePreviewFromBackend>
}

export type AllActions =
  SET_testIdAction
  | SET_contentAction
  | SET_displayNameAction
  | SET_displayIndexAction
  | SET_isSubmitTestAction
  | SET_weightAction
  | SET_testTypeIdAction
  | SET_editorTest_memoryLimitInKbAction
  | SET_editorTest_timeoutInMsAction
  | SET_editorTest_compileTimeoutInMsAction
  | SET_editorTest_maxDiskSpaceInKbAction
  | SET_editorTest_filesAction

  | ATTACH_testAsset_FULFILLEDAction
  | DETACH_testAsset_FULFILLEDAction


export function reducer(state: State = initial(), action: AllActions): State {

  switch (action.type) {

    case ActionType.SET_testId:
      return {
        ...state,
        id: action.id
      }

    case ActionType.SET_content:
      return {
        ...state,
        content: action.content
      }

    case ActionType.SET_displayName:
      return {
        ...state,
        displayName: action.displayName
      }

    case ActionType.SET_displayIndex:
      return {
        ...state,
        displayIndex: action.displayIndex
      }

    case ActionType.SET_isSubmitTest:
      return {
        ...state,
        isSubmitTest: action.isSubmitTest
      }

    case ActionType.SET_weight:
      return {
        ...state,
        weight: action.weight
      }

    case ActionType.SET_testTypeId:
      return {
        ...state,
        testTypeId: action.testTypeId
      }

    case ActionType.SET_editorTest_memoryLimitInKb:
      return {
        ...state,
        testSettings: {
          ...state.testSettings,
          memoryLimitInKb: action.memoryLimitInKb
        }
      }

    case ActionType.SET_editorTest_maxDiskSpaceInKb:
      return {
        ...state,
        testSettings: {
          ...state.testSettings,
          maxDiskSpaceInKb: action.maxDiskSpaceInKb
        }
      }

    case ActionType.SET_editorTest_timeoutInMs:
      return {
        ...state,
        testSettings: {
          ...state.testSettings,
          timeoutInMs: action.timeoutInMs
        }
      }
    case ActionType.SET_editorTest_compileTimeoutInMs:
      return {
        ...state,
        testSettings: {
          ...state.testSettings,
          compileTimeoutInMs: action.compileTimeoutInMs
        }
      }

    case ActionType.SET_editorTest_files:
      return {
        ...state,
        files: action.files
      }

    case EditorExerciseSiteActionTypes.ATTACH_testAsset_FULFILLED:
      return {
        ...state,
        files: state.files.concat(action.payload)
      }
    case EditorExerciseSiteActionTypes.DETACH_testAsset_FULFILLED:
      return {
        ...state,
        files: state.files.filter(p => p.id !== action.meta.assetPreviewId)
      }

    default:
      notExhaustive(action)
      return state
  }
}

